from dataclasses import dataclass, field
from .baseclass_dataclass import EmptyClass


@dataclass
class Entry_KanjiElement(EmptyClass):
    value: str  # keb
    info: field(default_factory=list[str])  # ke_inf*
    priority: field(default_factory=list[str])  # ke_pri*


@dataclass
class Entry_ReadingElement(EmptyClass):
    value: str  # reb
    no_kanji: bool  # re_nokanji?
    # reading is applied only to a set of KanjiElements
    atk: field(default_factory=list[str])  # re_restr*
    info: field(default_factory=list[str])  # re_inf*
    priority: field(default_factory=list[str])  # re_pri*


@dataclass
class Entry_SourceLanguage(EmptyClass):
    value: str  # lsource
    v_type: str | None  # lsource ls_type
    language: str | None  # lsource xml:lang
    wasei: str | None  # lsource ls_wasei


@dataclass
class Entry_Sense_Gloss(EmptyClass):
    value: str  # gloss
    v_type: str | None  # gloss g_type
    language: str | None  # gloss xml:lang
    wasei: str | None  # gloss g_gend


# @dataclass
# class Entry_Example(EmptyClass):
#     text: str  # ex_text
#     source: str  # ex_srce
#     language: str  # ex_sent xml:lang
#     sent: list[str]  # ex_sent
#     type: str  # ex_srce exsrc_type


@dataclass
class Entry_Sense(EmptyClass):
    # Sense is applied only to a set of Entry_KanjiElement
    aptk: field(default_factory=list[str])  # stagk*
    # Sense is applied only to a set of Entry_ReadingElement
    aptr: field(default_factory=list[str])  # stagr*
    part_of_speech: field(default_factory=list[str])  # pos*
    xref: field(default_factory=list[str])  # xref*
    antonym: field(default_factory=list[str])  # ant*
    use_field: field(default_factory=list[str])  # field*
    miscellaneous: field(default_factory=list[str])  # misc*
    info: field(default_factory=list[str])  # s_inf*
    source_language: field(default_factory=list[Entry_SourceLanguage])  # lsource*
    dialect: field(default_factory=list[str])  # dial*
    gloss: field(default_factory=list[Entry_Sense_Gloss])  # gloss*
    # example: field(default_factory=list[Entry_Example])  # example*


@dataclass
class Entry(EmptyClass):
    entry_sequence: int  # ent_seq
    kanji_element: field(default_factory=list[Entry_KanjiElement])  # k_ele*
    reading_element: field(default_factory=list[Entry_ReadingElement])  # r_ele+
    sense: field(default_factory=list[Entry_Sense])  # sense+




from dataclasses import dataclass, fields, field, is_dataclass


def _get_children_status(class_name, item):
    status = []
    if is_dataclass(item):
        if item.empty_status:
            return [f'{class_name} -> {item.empty_status}']
        else:
            return [f'{class_name} -> {x}' for x in item.children_status]

    if type(item) is list:
        for sub_item in item:
            status.extend(
                _get_children_status(class_name, sub_item)
            )

    return status


@dataclass
class EmptyClass:
    def __post_init__(self):
        self.class_name = self._set_class_name()
        self.empty_status = self._check_is_empty()
        self.children_status = self._check_children_status()

    @classmethod
    def _set_class_name(cls) -> str:
        return cls.__name__

    def _iterate_over(self):
        class_fields = [x for x in fields(self.__class__)
                        if x not in ['class_name', 'empty_status', 'children_status']]
        for class_field in class_fields:
            item = getattr(self, class_field.name)
            yield item

    def _check_is_empty(self) -> None | str:
        for item in self._iterate_over():
            if item:
                return None
        return f'{self.class_name} is empty!'

    def _check_children_status(self) -> (bool, list):
        children_status = []
        for item in self._iterate_over():
            children_status.extend(
                _get_children_status(self.class_name, item)
            )

        return children_status





from functools import wraps
from dataclasses import dataclass


def get_empty_error(func=None) -> (dataclass, str | None):
    @wraps(func)
    def wrapper(*args, **kwargs):
        datacls = func(*args, **kwargs)
        if datacls.is_empty:
            return (datacls, type(datacls).__name__)
        return (datacls, None)
    return wrapper



from .parsing_dataclass import Entry, Entry_Sense, Entry_Sense_Gloss, Entry_SourceLanguage, \
    Entry_KanjiElement, Entry_ReadingElement

JMDICT_OFFSET = 1


def process_jmdict_node(node):
    xml_1998_namespace = "{http://www.w3.org/XML/1998/namespace}"

    # ent_seq
    entry_sequence = node.find("./ent_seq").text

    # k_ele*
    kanji_element = []
    for elem in node.findall("./k_ele"):
        value = elem.find("./keb").text

        info = []
        for sub_elem in elem.findall("./ke_inf"):
            info.append(
                sub_elem.text
            )

        priority = []
        for sub_elem in elem.findall("./ke_pri"):
            priority.append(
                sub_elem.text
            )

        kanji_element.append(
            Entry_KanjiElement(
                value=value,
                info=info,
                priority=priority,
            )
        )

    # r_ele+
    reading_element = []
    for elem in node.findall("./r_ele"):
        value = elem.find("./reb").text
        no_kanji = True if elem.find("./re_nokanji") else False

        atk = []
        for sub_elem in elem.findall("./re_restr"):
            atk.append(
                sub_elem.text
            )

        info = []
        for sub_elem in elem.findall("./re_inf"):
            info.append(
                sub_elem.text
            )

        priority = []
        for sub_elem in elem.findall("./re_pri"):
            priority.append(
                sub_elem.text
            )

        reading_element.append(
            Entry_ReadingElement(
                value=value,
                no_kanji=no_kanji,
                atk=atk,
                info=info,
                priority=priority,
            )
        )

    # sense+
    sense = []
    for elem in node.findall("./sense"):
        aptk = []
        for sub_elem in elem.findall("./stagk"):
            aptk.append(
                sub_elem.text
            )

        aptr = []
        for sub_elem in elem.findall("./stagr"):
            aptr.append(
                sub_elem.text
            )

        part_of_speech = []
        for sub_elem in elem.findall("./pos"):
            part_of_speech.append(
                sub_elem.text
            )

        xref = []
        for sub_elem in elem.findall("./xref"):
            xref.append(
                sub_elem.text
            )

        antonym = []
        for sub_elem in elem.findall("./ant"):
            antonym.append(
                sub_elem.text
            )

        use_field = []
        for sub_elem in elem.findall("./field"):
            use_field.append(
                sub_elem.text
            )

        miscellaneous = []
        for sub_elem in elem.findall("./misc"):
            miscellaneous.append(
                sub_elem.text
            )

        info = []
        for sub_elem in elem.findall("./s_inf"):
            info.append(
                sub_elem.text
            )

        source_language = []
        for sub_elem in elem.findall("./lsource"):
            value = sub_elem.text
            v_type = sub_elem.get('ls_type', default=None)
            language = sub_elem.get(f'{xml_1998_namespace}lang', default=None)
            wasei = sub_elem.get('ls_wasei', default=None)

            source_language.append(
                Entry_SourceLanguage(
                    value=value,
                    v_type=v_type,
                    language=language,
                    wasei=wasei,
                )
            )

        dialect = []
        for sub_elem in elem.findall("./dial"):
            dialect.append(
                sub_elem.text
            )

        gloss = []
        for sub_elem in elem.findall("./gloss"):
            value = sub_elem.text
            v_type = sub_elem.get('g_type', default=None)
            language = sub_elem.get(f'{xml_1998_namespace}lang', default=None)
            wasei = sub_elem.get('g_gend', default=None)

            gloss.append(
                Entry_Sense_Gloss(
                    value=value,
                    v_type=v_type,
                    language=language,
                    wasei=wasei,
                )
            )

        # NO EXAMPLES FOUND IN JMDICT (Version 1.0 / Rev 1.09)
        # example = []
        # for sub_elem in elem.findall("./example"):
        #     text = sub_elem.text
        #     source = sub_elem.get('g_type', default=None)
        #     language = sub_elem.get('xml:lang', default=None)
        #     sent = []
        #     for sub2_elem in elem.findall("./sent"):
        #         sent.append(
        #             sub2_elem.text
        #         )
        #
        #     example.append(
        #         Entry_Example(
        #             text=text,
        #             source=source,
        #             language=language,
        #             logger=logger
        #         )
        #     )

        sense.append(
            Entry_Sense(
                aptk=aptk,
                aptr=aptr,
                part_of_speech=part_of_speech,
                xref=xref,
                antonym=antonym,
                use_field=use_field,
                miscellaneous=miscellaneous,
                info=info,
                source_language=source_language,
                dialect=dialect,
                gloss=gloss,
            )
        )

    entry = Entry(
        entry_sequence=entry_sequence,
        kanji_element=kanji_element,
        reading_element=reading_element,
        sense=sense,
    )

    return entry



from .kanjidic2_dataclass import Kanji_CharacterStandard, Kanji_Radical, Kanji_Description_Variant, Kanji_RNM_group, \
    Kanji_Description, Kanji_DictionaryNumber, Kanji_QueryCode, Kanji_RNM_Reading, Kanji_RNM_Meaning, Kanji_RNM, Kanji
from .jmdict_dataclass import Entry, Entry_Sense, Entry_Sense_Gloss, Entry_SourceLanguage, \
    Entry_KanjiElement, Entry_ReadingElement
from .get_error_wrapper import get_empty_error



from dataclasses import dataclass, field, fields
from .baseclass_dataclass import EmptyClass


@dataclass
class Kanji_CharacterStandard(EmptyClass):
    value: str  # cp_value
    v_type: str  # cp_type


@dataclass
class Kanji_Radical(EmptyClass):
    value: str  # rad_value
    v_type: str  # rad_type


@dataclass
class Kanji_Description_Variant(EmptyClass):
    value: str  # variant
    v_type: str  # var_type


@dataclass
class Kanji_Description(EmptyClass):
    grade: int | None  # grade?
    stroke_count: field(default_factory=list[int])  # stroke_count+
    variant: field(default_factory=list[Kanji_Description_Variant])  # variant*
    frequency: int | None  # freq?
    radical_name: field(default_factory=list[str])  # rad_name*
    jlpt: int | None  # jlpt?


@dataclass
class Kanji_DictionaryNumber(EmptyClass):
    value: str  # dic_ref
    v_type: str  # dr_type
    volume: int | None  # m_vol
    page: int | None  # m_page


@dataclass
class Kanji_QueryCode(EmptyClass):
    value: str  # q_code
    v_type: str  # qc_type
    misclassification: str | None  # skip_misclass


@dataclass
class Kanji_RNM_Reading(EmptyClass):
    value: str  # reading
    v_type: str  # r_type #REQUIRED
    status: str | None  # r_status


@dataclass
class Kanji_RNM_Meaning(EmptyClass):
    value: str  # meaning
    v_type: str  # m_lang #IMPLIED


@dataclass
class Kanji_RNM_group(EmptyClass):
    reading: list[Kanji_RNM_Reading]  # reading*
    meaning: list[Kanji_RNM_Meaning]  # meaning*


@dataclass
class Kanji_RNM(EmptyClass):
    rmn_group: field(default_factory=list[Kanji_RNM_group])  # rmgroup*
    names_readings: field(default_factory=list[str])  # nanori*


@dataclass
class Kanji(EmptyClass):
    """Kanji class to describe kanji from kanjidic2"""
    kanji: str  # literal
    standard: field(default_factory=list[Kanji_CharacterStandard])  # codepoint (cp_value+)
    radical: field(default_factory=list[Kanji_Radical])  # radical (rad_value+)
    description: Kanji_Description  # misc (grade?, stroke_count+, variant*, freq?, rad_name*,jlpt?)
    dictionary_number: field(default_factory=list[Kanji_DictionaryNumber])  # dic_number?
    query_code: field(default_factory=list[Kanji_QueryCode])  # query_code? (q_code+)
    rnm: field(default_factory=list[Kanji_RNM])  # reading_meaning? (rmgroup*, nanori*)



from .helper import extract_text
from .parsing_dataclass import Kanji_CharacterStandard, Kanji_Radical, Kanji_Description_Variant, Kanji_RNM_group, \
    Kanji_Description, Kanji_DictionaryNumber, Kanji_QueryCode, Kanji_RNM_Reading, Kanji_RNM_Meaning, Kanji_RNM, Kanji

KANJIDIC_OFFSET = 1


def process_kanjidic_node(node):
    # kanji
    kanji = node.find("./literal").text

    # codepoint (cp_value+)
    standard = []
    for elem in node.findall("./codepoint/cp_value"):
        standard.append(
            Kanji_CharacterStandard(
                v_type=elem.get('cp_type'),
                value=elem.text,
            )
        )

    # radical (rad_value+)
    radical = []
    for elem in node.findall("./radical/rad_value"):
        radical.append(
            Kanji_Radical(
                v_type=elem.get('rad_type'),
                value=elem.text,
            )
        )

    # misc (grade?, stroke_count+, variant*, freq?, rad_name*,jlpt?)
    grade = extract_text(node.find("./misc/grade"))

    stroke_count = []
    for elem in node.findall("./misc/stroke_count"):
        stroke_count.append(
            elem.text
        )

    variant = []
    for elem in node.findall("./misc/variant"):
        variant.append(
            Kanji_Description_Variant(
                v_type=elem.get('var_type'),
                value=elem.text,
            )
        )

    frequency = extract_text(node.find("./misc/freq"))

    radical_name = []
    for elem in node.findall("./misc/rad_name"):
        radical_name.append(
            elem.text
        )

    jlpt = extract_text(node.find("./misc/jlpt"))

    description = Kanji_Description(
        grade=grade,
        stroke_count=stroke_count,
        variant=variant,
        frequency=frequency,
        radical_name=radical_name,
        jlpt=jlpt,
    )

    # dic_number?
    dictionary_reference = []
    for elem in node.findall("./dic_number/dic_ref"):
        dictionary_reference.append(
            Kanji_DictionaryNumber(
                value=extract_text(elem),
                v_type=elem.get('dr_type'),
                volume=elem.get('m_vol', default=None),
                page=elem.get('m_page', default=None),
            )
        )

    # query_code? (q_code+)
    query_code = []
    for elem in node.findall("./query_code/q_code"):
        query_code.append(
            Kanji_QueryCode(
                value=extract_text(elem),
                v_type=elem.get('qc_type'),
                misclassification=elem.get('skip_misclass', default=None),
            )
        )

    # reading_meaning? (rmgroup*, nanori*)
    rnm = []
    for elem in node.findall("./reading_meaning"):
        rnm_group = []
        for sub_elem in elem.findall("./rmgroup"):
            reading = []
            for sub2_elem in sub_elem.findall("./reading"):
                reading.append(
                Kanji_RNM_Reading(
                    value=extract_text(sub2_elem),
                    v_type=sub2_elem.get('r_type', default=None),
                    status=sub2_elem.get('status', default=None),
                )
                )

            meaning = []
            for sub2_elem in sub_elem.findall("./meaning"):
                meaning.append(
                    Kanji_RNM_Meaning(
                        value=extract_text(sub2_elem),
                        v_type=sub2_elem.get('m_lang', default=None),
                    )
                )

            rnm_group.append(
                Kanji_RNM_group(
                    reading=reading,
                    meaning=meaning,
                )
            )

        nanori = []
        for sub_elem in elem.findall("./nanori"):
            nanori.append(
                sub_elem.text
            )

        rnm.append(
            Kanji_RNM(
                rmn_group=rnm_group,
                names_readings=nanori,
            )

        )

    entry = Kanji(
        kanji=kanji,
        standard=standard,
        radical=radical,
        description=description,
        dictionary_number=dictionary_reference,
        query_code=query_code,
        rnm=rnm,
    )

    return entry



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['japanese_dictionary_parser',
 'japanese_dictionary_parser.parsing_utils',
 'japanese_dictionary_parser.parsing_utils.parsing_dataclass']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'japanese-dictionary-parser',
    'version': '0.1.1',
    'description': 'Parsers for japanese dictionaries: JMdict and kanjidic2',
    'long_description': '# Japanese dictionary parser\n___\nThe package contains two classes for parsing dictionaries:\n* JMdict: contains information on japanese vocabulary\n* kanjidic2: contains information on japanese kanji\n',
    'author': 'Maxim R',
    'author_email': 'joey-hzpfywtd@hotmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



from .parsing_utils import process_kanjidic_node, process_jmdict_node, ParsingBase


class KanjidicParser(ParsingBase):
    def _modify_offset(self):
        self.offset = 1

    def _parse_elem(self, elem, logger):
        return process_kanjidic_node(elem)


class JmdictParser(ParsingBase):
    def _modify_offset(self):
        self.offset = 0

    def _parse_elem(self, elem, logger):
        return process_jmdict_node(elem)



from .parsing_dataclass import Entry, Kanji
import xml.etree.ElementTree as ET
import logging
import pathlib


def print_node(node) -> None:
    for child_node in ET.tostring(node).decode('utf-8').split('\n'):
        if child_node:
            print(child_node)
    return


def get_text(node, xpath):
    text = node.find(xpath)
    if text:
        return text.text
    return ''


def extract_text(node, default=None) -> str | None:
    if node is not None:
        return node.text
    else:
        return default


def get_all_text(node, xpath):
    all_text = []
    for child_node in node.findall(xpath):
        if child_node.text:
            all_text.append(
                child_node.text
            )
    return all_text


def get_node_names(xml_node):
    name = xml_node.tag
    attributes = '__'.join((f'{k}={v}' for k, v in xml_node.attrib.items()))
    return f'{name}___{attributes}'


def is_child_node(xml_node, xpath: str) -> bool:
    return True if xml_node.find(xpath) is not None else False


def add_to_dict_if_exist(dictionary: dict, key: str, xml_node, xpath: str) -> None:
    if is_child_node(xml_node, xpath):
        dictionary[key] = get_text(xml_node, xpath)


def parse_xml(path, parsing_func, offset: int = 0, logger=None, test=False):
    tree = ET.parse(path)
    root = tree.getroot()

    elements = []
    for elem in root[offset:]:
        elements.append(
            parsing_func(elem, logger)
        )
        if test:
            return elements.pop()
    return elements


def _log(log_parsing: bool, logger: logging, _method: str, message: str | Exception):
    if log_parsing:
        if logger:
            if _method == 'critical':
                logging.critical(message, exc_info=True)
            else:
                _log_method = getattr(logging, _method)
                _log_method(message)
        else:
            print(message)


class ParsingBase:
    def __init__(self, path: str | pathlib.Path, log_parsing: bool = True, logger: logging = None):
        self.path = path
        self.log_parsing = log_parsing
        self.logger = logger
        self.offset = None
        self._modify_offset()

    def _parse_elem(self, elem, logger) -> Entry | Kanji:
        pass

    def _modify_offset(self) -> None:
        pass

    def __iter__(self):
        tree = ET.parse(self.path)
        root = tree.getroot()

        for idx, elem in enumerate(root[self.offset:]):
            _log(self.log_parsing, self.logger, 'info', f'Parsing element - {idx}')
            try:
                entry = self._parse_elem(elem, self.logger)
                item_status = entry.children_status
                for _status in item_status:
                    _log(self.log_parsing, self.logger, 'info', _status)
                _log(self.log_parsing, self.logger, 'info', f'Parsed element: {entry}')
            except Exception as e:
                entry = None
                _log(self.log_parsing, self.logger, 'critical', e)

            yield entry

